from __future__ import annotations

import typing as t

import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from tktl.core import ExtendedEnum


class EndpointKinds(str, ExtendedEnum):
    ADVANCED = "advanced"
    TYPED = "typed"
    ARROW = "arrow"
    PROFILED = "profiled"


class ProfileKinds(str, ExtendedEnum):
    BINARY = "binary"
    REGRESSION = "regression"


class ArrowFormatKinds(str, ExtendedEnum):
    DATAFRAME = "dataframe"
    SERIES = "series"
    ARRAY = "array"

    @staticmethod
    def from_type(t: t.Union[pd.DataFrame, pd.Series, np.ndarray]) -> ArrowFormatKinds:
        if isinstance(t, pd.DataFrame):
            return ArrowFormatKinds.DATAFRAME
        if isinstance(t, pd.Series):
            return ArrowFormatKinds.SERIES
        if isinstance(t, np.ndarray):
            return ArrowFormatKinds.ARRAY
        raise ValueError(f"{type(t)} is not an arrow communication type")



import typing as t

import pandas as pd  # type: ignore

from tktl.core.loggers import LOG

JSONObject = t.Dict[t.AnyStr, t.Any]
JSONArray = t.List[t.Any]
JSONStructure = t.Union[JSONArray, JSONObject]


def coerce_dataframe_to_series(df: pd.DataFrame) -> t.Union[pd.DataFrame, pd.Series]:
    if len(df.columns) == 1:
        # Since we store dataframes in parquet files, we assume that a
        # one column dataframe used to be a Series.
        return df.iloc[:, 0]
    return df


def coerce_series_to_dataframe(obj: t.Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
    if isinstance(obj, pd.Series):
        return pd.DataFrame({str(obj.name): obj})
    if len(obj.columns) == 1:
        LOG.error(
            "Storing 1 column dataframe, this will become a Series on the way back"
        )
    return obj



import json
import logging
import os
from typing import Dict, Optional, Union

import yaml

from tktl import __version__
from tktl.core.config import Settings, settings
from tktl.core.managers.constants import CONFIG_FILE_TEMPLATE

log = logging.getLogger("root")


class BaseConfigManager(object):
    """Base class for managing a configuration file.
    Based on the amazing github.com/polyaxon/polyaxon config manager setup
    """

    TKTL_DIR: Optional[str]
    CONFIG_FILE_NAME: str
    CONFIG: Optional[dict]
    SETTINGS: Settings = settings

    @classmethod
    def get_tktl_config_path(cls):
        if not os.path.exists(cls.SETTINGS.TKTL_CONFIG_PATH):
            cls.create_dir(cls.SETTINGS.TKTL_CONFIG_PATH)
        return cls.SETTINGS.TKTL_CONFIG_PATH

    @classmethod
    def get_tktl_config(cls, init=False):
        path = os.path.join(cls.get_tktl_config_path(), cls.SETTINGS.CONFIG_FILE_NAME)
        if os.path.exists(path):
            with open(path, "r") as cfg:
                return json.load(cfg)
        else:
            if init:
                with open(path, "w") as cfg:
                    json.dump({}, cfg)
            return {}

    @classmethod
    def set_tktl_config(cls, values: Dict):
        path = cls.get_tktl_config_path()
        cfg_path = os.path.join(path, cls.SETTINGS.CONFIG_FILE_NAME)
        with open(cfg_path, "w") as cfg:
            json.dump(values, cfg, indent=2, sort_keys=True)
            cfg.write("\n")
        return values

    @classmethod
    def clear_tktl_config(cls):
        cls.set_tktl_config(values={})

    @staticmethod
    def create_dir(dir_path: str) -> None:
        if not os.path.exists(dir_path):
            try:
                os.makedirs(dir_path)
            except OSError as e:
                # Except permission denied and potential race conditions
                # in multi-threaded environments.
                log.error(f"Could not create config directory {dir_path}: {repr(e)}")

    @classmethod
    def get_config_file_path(cls) -> str:
        if not cls.TKTL_DIR:
            # local to this directory
            base_path = os.path.join(".")
        else:
            base_path = cls.TKTL_DIR
        return os.path.join(base_path, cls.CONFIG_FILE_NAME)

    @classmethod
    def init_config(cls, init: Union[str, bool] = False) -> None:
        cls.set_config(init=init)

    @classmethod
    def is_initialized(cls) -> bool:
        config_file_path = cls.get_config_file_path()
        return os.path.isfile(config_file_path)

    @classmethod
    def set_config(cls, init: Union[str, bool] = False) -> None:
        config_file_path = cls.get_config_file_path()

        if os.path.isfile(config_file_path) and init:
            log.debug(
                "%s file already present at %s", cls.CONFIG_FILE_NAME, config_file_path
            )
            return

        if init:
            config_file = CONFIG_FILE_TEMPLATE.format(version=__version__)
            with open(config_file_path, "w") as config_file_handle:
                config_file_handle.write(config_file)

    @classmethod
    def get_config(cls) -> Optional[dict]:
        if not cls.is_initialized():
            return None
        config_file_path = cls.get_config_file_path()
        with open(config_file_path, "r") as config_file:
            return yaml.safe_load(config_file)

    @classmethod
    def get_value(cls, key: str) -> Union[Optional[str], Optional[dict]]:
        config = cls.get_config()
        if config:
            if key in config.keys():
                return config[key]
            else:
                log.warning("Config `%s` has no key `%s`", cls.CONFIG_FILE_NAME, key)
        return None



import pathlib

import click

from tktl.cli.common import ClickCommand
from tktl.commands.lazify import lazify_file
from tktl.core.config import settings


@click.command(
    "lazify",
    help="Add metadata to parquet file for lazy loading",
    cls=ClickCommand,
    **settings.HELP_COLORS_DICT,
)
@click.argument("source_path")
@click.argument("target_path")
@click.pass_context
def lazify(ctx, source_path: str, target_path: str):

    source_pathlib_path = pathlib.Path(source_path)
    target_pathlib_path = pathlib.Path(target_path)

    if not source_pathlib_path.is_file():
        raise click.BadParameter(
            f"'{source_path}' does not exist or is not a file", param_hint="source_path"
        )
    if source_pathlib_path.suffix != ".pqt":
        raise click.BadParameter(
            f"'{source_path}' does not seem to be a parquet file",
            param_hint="source_path",
        )

    lazify_file(
        source_path=source_pathlib_path, target_path=target_pathlib_path, data=None
    )



from fastapi import BackgroundTasks  # noqa: F401

from tktl.core.clients.future.arrow import ArrowClient  # noqa: F401
from tktl.core.clients.future.rest import RestClient  # noqa: F401

from .registration.decorator import Tktl  # noqa: F401
from .registration.endpoints import (  # noqa: F401
    ArrowEndpoint,
    ProfiledEndpoint,
    TypedEndpoint,
)



import typing as t
from functools import singledispatch

import numpy as np
import pandas as pd  # type: ignore

from tktl.future.registration.exceptions import SerializationError
from tktl.future.utils import JSONStructure

from .dataframe import (  # noqa: F401
    dataframe_deserialize,
    dataframe_serialize,
    dataframe_to_example,
)
from .numpy import numpy_deserialize, numpy_serialize, numpy_to_example  # noqa: F401
from .series import (  # noqa: F401
    series_deserialize,
    series_serialize,
    series_to_example,
)


@singledispatch
def serialize(value: t.Union[np.ndarray, pd.Series, pd.DataFrame]):
    raise SerializationError(f"Can't serialize value of type {type(value)}")


@serialize.register
def _serialize_numpy(value: np.ndarray):
    return numpy_serialize(value=value)


@serialize.register
def _serialize_series(value: pd.Series):
    return series_serialize(value=value)


@serialize.register
def _serialize_dataframe(value: pd.DataFrame):
    return dataframe_serialize(value=value)


@singledispatch
def deserialize(
    sample: t.Union[np.ndarray, pd.Series, pd.DataFrame], *, value: JSONStructure,
):
    raise SerializationError(f"Can't deserialize value of type {type(sample)}")


@deserialize.register
def _deserialize_numpy(sample: np.ndarray, *, value: JSONStructure):
    return numpy_deserialize(value=value, sample=sample)


@deserialize.register
def _deserialize_series(sample: pd.Series, *, value: JSONStructure):
    return series_deserialize(value=value, sample=sample)


@deserialize.register
def _deserialize_dataframe(sample: pd.DataFrame, *, value: JSONStructure):
    return dataframe_deserialize(value=value, sample=sample)


@singledispatch
def to_example(value: t.Union[np.ndarray, pd.Series, pd.DataFrame]):
    raise SerializationError(f"Can't create example value of type {type(value)}")


@to_example.register
def _to_example_numpy(value: np.ndarray):
    return numpy_to_example(value=value)


@to_example.register
def _to_example_series(value: pd.Series):
    return series_to_example(value=value)


@to_example.register
def _to_example_dataframe(value: pd.DataFrame):
    return dataframe_to_example(value=value)



from .abc import Endpoint  # noqa: F401
from .advanced import AdvancedEndpoint  # noqa: F401
from .arrow import ArrowEndpoint  # noqa: F401
from .profiled import ProfiledEndpoint  # noqa: F401
from .typed import TypedEndpoint  # noqa: F401



import pathlib
import typing as t

from tktl.core.exceptions import EndpointInstantiationError

from .endpoints import AdvancedEndpoint, ArrowEndpoint, ProfiledEndpoint, TypedEndpoint
from .endpoints.abc import XType, YType

# Enpoint types sorted by precedence
ENDPOINT_TYPES = [ProfiledEndpoint, ArrowEndpoint, TypedEndpoint, AdvancedEndpoint]


class Tktl:
    def __init__(self):
        self._endpoints = []

    @property
    def endpoints(self):
        return self._endpoints

    def endpoint(
        self,
        *,
        X: t.Union[XType, str] = None,
        y: t.Union[YType, str] = None,
        profile: t.Optional[str] = None,
        profile_columns: t.Optional[t.List[str]] = None,
        **kwargs,
    ):

        if isinstance(X, str):
            X = pathlib.Path(X)

        if isinstance(y, str):
            y = pathlib.Path(y)

        try:
            Constructor = next(
                endpoint
                for endpoint in ENDPOINT_TYPES
                if endpoint.supported(X=X, y=y, profile=profile)
            )
        except StopIteration:
            raise EndpointInstantiationError(
                f"Arguments of type X={type(X)} y={type(y)} and profile={profile}"
                " are not supported by any endpoint type."
            )

        def decorator(f):
            endpoint = Constructor(
                name=f.__name__,
                func=f,
                X=X,
                y=y,
                profile_columns=profile_columns,
                profile=profile,
                **kwargs,
            )
            self._endpoints.append(endpoint)
            return f

        return decorator



import click
from taktile_types.enums.instance import ServiceType

from tktl.cli.common import ClickCommand
from tktl.commands.health import GetGrpcHealthCommand, GetRestHealthCommand
from tktl.core.config import settings
from tktl.core.exceptions import TaktileSdkError
from tktl.core.exceptions.exceptions import APIClientException
from tktl.core.loggers import LOG
from tktl.core.managers.auth import AuthConfigManager
from tktl.login import login, logout


@click.command(
    "health",
    help="Health check your endpoints",
    cls=ClickCommand,
    **settings.HELP_COLORS_DICT
)
@click.option("-r", "--repo", help="Repository owner & name: owner/repo-name")
@click.option("-b", "--branch", help="Branch name")
@click.option(
    "-s",
    "--service",
    help="Service kind",
    type=click.Choice([ServiceType.REST, ServiceType.GRPC]),
    default=ServiceType.REST,
)
@click.option(
    "-l", "--local", help="Run against local endpoint", is_flag=True, default=False
)
@click.pass_context
def health(
    ctx, repo: str, branch: str, service: str = ServiceType.REST, local: bool = False,
):
    if not local:
        if not repo and not branch:
            return click.echo(health.get_help(ctx=ctx))
        if (repo and not branch) or (branch and not repo):
            LOG.error("If not running locally, must set branch and repo name")
            return

    api_key = AuthConfigManager.get_api_key()
    if local and not api_key:
        login(api_key="NOT SET")
    command = (
        GetRestHealthCommand(
            repository=repo, branch_name=branch, local=local, skip_auth=True
        )
        if service == ServiceType.REST
        else GetGrpcHealthCommand(
            repository=repo, branch_name=branch, local=local, skip_auth=True
        )
    )
    try:
        command.execute()
    except APIClientException:
        exit(1)
    except TaktileSdkError as e:
        LOG.error(str(e))
    finally:
        if not api_key:
            logout()



from typing import Optional

from pydantic import UUID4, BaseModel


class DeploymentBase(BaseModel):
    id: Optional[UUID4] = None
    git_branch: Optional[str] = None
    git_hash: Optional[str] = None
    requester_id: Optional[UUID4] = None
    repository_id: Optional[UUID4] = None
    repo_ref_id: int
    status_id: Optional[UUID4] = None
    ecr_repo_url: str



import json
from io import BytesIO
from typing import Any, List, Sequence, Tuple, Type, Union

import numpy  # type: ignore
import pandas  # type: ignore
import pyarrow  # type: ignore
from pandas.core.dtypes.inference import is_sequence  # type: ignore

from tktl.core.exceptions import UnsupportedInputTypeException

from ..base import ObjectSerializer
from ..utils import is_valid_list_of_lists
from .codecs import (
    DEFAULT_SERIALIZED_COLUMN_NAME,
    DESERIALIZE_SEQUENCE_OPS,
    ArbitraryBinaryArray,
)
from .utils import (
    _get_final_serializer,
    _serialize_dicts_or_list_of_dicts,
    add_deserializing_metadata,
    inject_metadata,
)


class DataFrameSerializer(ObjectSerializer):
    @classmethod
    def deserialize(cls, value: pyarrow.Table) -> pandas.DataFrame:
        return value.to_pandas()

    @classmethod
    @inject_metadata
    def serialize(cls, value: pandas.DataFrame) -> pyarrow.Table:
        table = pyarrow.Table.from_pandas(value, preserve_index=True)
        return table


class SeriesSerializer(ObjectSerializer):
    @classmethod
    def find_idx(cls, table: pyarrow.Table):
        return sorted(
            [col for col in table.column_names if col.startswith("__index_level_")],
            key=lambda x: int("".join(s for s in x if s.isdigit())),
        )

    @classmethod
    def deserialize(cls, value: pyarrow.Table) -> pandas.Series:
        idxs = cls.find_idx(value)
        other_names = [c for c in value.column_names if c not in idxs]
        assert len(other_names) > 0
        name = other_names[0]
        if len(idxs) > 1:
            idx_values = [value.column(i).to_pandas().values for i in idxs]
            series_index = pandas.MultiIndex.from_arrays(idx_values)
        elif len(idxs) == 1:
            series_index = pandas.Index(value.column(idxs[0]).to_pandas().values)
        else:
            series_index = None
        return pandas.Series(
            value.column(name).to_pandas().values, index=series_index, name=name
        )

    @classmethod
    @inject_metadata
    def serialize(cls, value: pandas.Series) -> pyarrow.Table:
        as_df = pandas.DataFrame(value)
        tbl = pyarrow.Table.from_pandas(as_df, preserve_index=True)
        return tbl


class ArraySerializer(ObjectSerializer):
    @classmethod
    def deserialize(cls, value: pyarrow.Table):
        if len(value.columns) == 1:
            return value.column(0).to_pandas().to_numpy()
        else:
            return numpy.concatenate(
                [
                    value.column(i).to_pandas().to_numpy().reshape(-1, 1)
                    for i in range(len(value.columns))
                ],
                axis=1,
            )

    @classmethod
    @inject_metadata
    def serialize(cls, value: numpy.ndarray, names: List[str] = None) -> pyarrow.Table:
        arrow_arrays = []
        if value.ndim == 1:
            c = 1
            arrow_arrays = [pyarrow.array(value)]
        elif value.ndim == 2:
            r, c = value.shape
            for column in range(c):
                arrow_arrays.append(pyarrow.array(value[:, column]))
        else:
            raise UnsupportedInputTypeException("Must serialise using binary format")
        if not names:
            names = [str(i) for i in range(c)]
        return pyarrow.Table.from_arrays(arrays=arrow_arrays, names=names)


class SequenceSerializer(ObjectSerializer):
    @classmethod
    def deserialize(cls, value: pyarrow.Table) -> Sequence:
        pipeline = json.loads(value.schema.metadata[b"DESERIALIZERS"].decode("utf-8"))
        for function_ref in pipeline:
            value = DESERIALIZE_SEQUENCE_OPS[function_ref](value)
        return value

    @staticmethod
    def _serialize(value: Sequence) -> Tuple[Union[pyarrow.Array, pyarrow.Table], Type]:
        if isinstance(value, list) or isinstance(value, tuple):
            if len(value) > 0:
                first = value[0]
                if isinstance(first, list):
                    is_valid = is_valid_list_of_lists(value)
                    if is_valid:
                        return pyarrow.array(value), list
                    raise UnsupportedInputTypeException(
                        "Must serialise using binary format"
                    )
                elif isinstance(first, dict):
                    return _serialize_dicts_or_list_of_dicts(value), list
                # more specific than numpy's is_sequence implementation
                elif not is_sequence(first):
                    return pyarrow.array(value), list
            else:
                raise ValueError("Invalid inputs provided: empty sequence")
        if isinstance(value, dict):
            return _serialize_dicts_or_list_of_dicts(value), list
        else:
            raise

    @classmethod
    @inject_metadata
    def serialize(cls, value: Sequence) -> pyarrow.Table:
        intermediate, final_type = cls._serialize(value)
        final_deserializer = _get_final_serializer(
            initial_type=type(value), final_type=final_type
        )
        if isinstance(intermediate, pyarrow.Table):
            deserializer_pipeline = ["NATIVE_dict"]
            if final_deserializer:
                deserializer_pipeline.append(final_deserializer)

            intermediate = add_deserializing_metadata(
                intermediate, deserializers=deserializer_pipeline
            )

        elif isinstance(intermediate, pyarrow.Array):
            as_table = pyarrow.Table.from_arrays(
                [intermediate], names=[DEFAULT_SERIALIZED_COLUMN_NAME]
            )
            deserializer_pipeline = ["SELECT_column", "NATIVE_list"]
            if final_deserializer:
                deserializer_pipeline.append(final_deserializer)
            intermediate = add_deserializing_metadata(
                as_table, deserializers=deserializer_pipeline
            )
        else:
            raise UnsupportedInputTypeException("Must serialize using binary format")
        return intermediate


class BinarySerializer(ObjectSerializer):
    @classmethod
    def deserialize(cls, value: pyarrow.Table):
        val = value.to_pandas()[DEFAULT_SERIALIZED_COLUMN_NAME].values
        if not val:
            return None
        mem_file = BytesIO(val[0])
        return numpy.load(mem_file, allow_pickle=True)  # noqa

    @classmethod
    @inject_metadata
    def serialize(cls, value: Any) -> pyarrow.Table:
        bin_schema = pyarrow.schema(
            [(DEFAULT_SERIALIZED_COLUMN_NAME, pyarrow.binary())]
        )
        arr = ArbitraryBinaryArray(val=value)
        deserializer_pipeline = ["SELECT_column"]
        intermediate = pyarrow.table([arr], schema=bin_schema)
        return add_deserializing_metadata(
            table=intermediate, deserializers=deserializer_pipeline
        )



import typing as t

from tktl.core.clients.future.http import API
from tktl.core.clients.future.taktile import DeploymentApiClient
from tktl.core.clients.http_client import interpret_response
from tktl.core.future.t import EndpointKinds, ProfileKinds
from tktl.core.loggers import LOG
from tktl.core.schemas.future.service import (
    EndpointInfoSchema,
    InfoEndpointResponseModel,
)
from tktl.core.schemas.repository import RepositoryDeployment, _format_http_url

from .endpoint import RestEndpoints


class RestClient:

    endpoints: RestEndpoints

    def __init__(
        self,
        api_key: str,
        repository_name: str,
        branch_name: str,
        url: t.Optional[str] = None,
    ):
        self._api_key = api_key
        self._repository_name = repository_name
        self._branch_name = branch_name
        self._url = RestClient.__get_url(
            url=url,
            repository_name=repository_name,
            branch_name=branch_name,
            api_key=api_key,
        )

        self._client = RestClient.__get_deployment_client(
            url=self._url, api_key=api_key
        )

        self._info = self.__get_info(api_key=api_key, client=self._client)

        self.endpoints = RestEndpoints(api=self._client, infos=self._info.endpoints)

    @staticmethod
    def __get_deployment(
        *, repository_name: str, branch_name: str, client: DeploymentApiClient
    ) -> RepositoryDeployment:
        return client.get_deployment_by_branch_name(
            repository_name=repository_name, branch_name=branch_name
        )

    @staticmethod
    def __get_info(*, api_key: str, client: API) -> InfoEndpointResponseModel:
        response = client.get("/info")
        response.raise_for_status()

        response_content = response.json()

        if "schema_version" not in response_content:

            # Support for classic deployments
            # TODO: Delete this in taktile-cli 1.0
            LOG.warning("Using compatibility mode for classic deployments...")

            to_profile_kind: t.Dict[str, ProfileKinds] = {
                "binary": ProfileKinds.BINARY,
                "regression": ProfileKinds.REGRESSION,
            }

            def to_kind(endpoint: t.Dict[str, t.Any]) -> EndpointKinds:
                if endpoint["kind"] in ["binary", "regression"]:
                    return EndpointKinds.PROFILED
                if endpoint["kind"] in ["custom"]:
                    return EndpointKinds.ADVANCED
                return EndpointKinds.ADVANCED

            return InfoEndpointResponseModel(
                schema_version="unknown",
                taktile_cli="unknown",
                profiling="unknown",
                endpoints=[
                    EndpointInfoSchema(
                        name=x["name"],
                        path=f"/model/{x['name']}",
                        kind=to_kind(x),
                        profile_kind=to_profile_kind.get(x["kind"], None),
                        explain_path=f"/model/{x['name']}/explain",
                        response_kind=None,
                        input_names=x["input_names"],
                        output_names=x["output_names"] if x["output_names"] else None,
                        profile_columns=x["profiling_columns"],
                        input_example=x["explain_input_example"],
                    )
                    for x in response_content
                ],
            )

        return interpret_response(response, InfoEndpointResponseModel)

    @staticmethod
    def __get_dapi_client(*, api_key: str) -> DeploymentApiClient:
        return DeploymentApiClient(api_key=api_key)

    @staticmethod
    def __get_deployment_client(*, url: str, api_key: str) -> API:
        client = API(api_url=url, headers={"X-Api-Key": api_key})

        return client

    @staticmethod
    def __get_url(
        *, url: t.Optional[str], repository_name: str, branch_name: str, api_key: str
    ) -> str:
        if url:
            return url

        client = RestClient.__get_dapi_client(api_key=api_key)
        deployment = RestClient.__get_deployment(
            repository_name=repository_name, branch_name=branch_name, client=client
        )

        return _format_http_url(deployment.public_docs_url, docs=False)



import http
from typing import List

from pydantic import ValidationError


class TklException(Exception):
    ...


class MissingDocker(Exception):
    ...


class NotInitializedError(Exception):
    ...


class VersionError(Exception):
    ...


class CLIError(Exception):
    ...


class ModelNotFoundError(Exception):
    ...


class EndpointException(Exception):
    ...


class ConversionException(EndpointException):
    ...


class ValidationException(EndpointException):
    ...


class ProjectValidationException(TklException):
    ...


class HTTPException(Exception):
    def __init__(self, status_code: int, detail: str = None) -> None:
        if detail is None:
            detail = http.HTTPStatus(status_code).phrase
        self.status_code = status_code
        self.detail = detail

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.detail!r})"


class APISwitchException(Exception):
    # TODO: Remove this in taktile-cli 1.0
    pass


class APIClientException(HTTPException):
    ...

    def __str__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.detail!r})"


class APIClientExceptionRetryable(APIClientException):
    ...


class UserRepoValidationException(ProjectValidationException):
    def __init__(
        self, missing_files: List, missing_directories: List, missing_config: bool
    ):
        self.missing_files = missing_files
        self.missing_directories = missing_directories
        self.missing_config = missing_config


class NoContentsFoundException(UserRepoValidationException):
    ...


class TktlConfigInvalidException(ProjectValidationException):
    def __init__(self, validation_error: ValidationError):
        self.err = validation_error


def validate_config(fn):
    from tktl.core.managers.project import ProjectManager

    def wrapper(*args, **kwargs):
        if ProjectManager.get_config() is None:
            raise CLIError(
                "No configuration found. Are you sure you have a tktl.yaml file in "
                "your current directory? Run tktl init to start a new project"
            )
        return fn(*args, **kwargs)

    return wrapper



from tktl.core import ExtendedEnum


class UserRepoConfigFileNameT(str, ExtendedEnum):
    YAML = "tktl.yaml"
    YML = "tktl.yml"


class UserProjectFileT(str, ExtendedEnum):
    FILE = "file"
    DIRECTORY = "dir"


class RequiredUserProjectPathsT(str, ExtendedEnum):
    SRC = "src"
    ASSETS = "assets"
    TESTS = "tests"
    TKTL_YML = "tktl.yml"
    TKTL_YAML = "tktl.yaml"
    DOCKERFILE = ".buildfile"
    REQS = "requirements.txt"

    @classmethod
    def strictly_required_files(cls):
        return {cls.DOCKERFILE.value, cls.REQS.value}

    @classmethod
    def strictly_required_dirs(cls):
        return {cls.SRC.value, cls.ASSETS.value, cls.TESTS.value}


class DeploymentStateFailed(str, ExtendedEnum):
    BUILDING_FAILED = "building_failed"
    DEPLOYING_FAILED = "deploying_failed"
    ROUTING_FAILED = "routing_failed"
    PROFILING_FAILED = "profiling_failed"
    ERRORED = "errored"


class DeploymentStateSucceed(str, ExtendedEnum):
    RUNNING = "running"


class DeploymentStatePending(str, ExtendedEnum):
    BUILDING = "building"
    DEPLOYING = "deploying"
    ROUTING = "routing"
    PROFILING = "profiling"


class DeploymentStatesT(str, ExtendedEnum):

    BUILDING = "building"
    DEPLOYING = "deploying"
    ROUTING = "routing"
    PROFILING = "profiling"
    RUNNING = "running"
    DELETED = "deleted"

    BUILDING_FAILED = "building_failed"
    DEPLOYING_FAILED = "deploying_failed"

    ROUTING_FAILED = "routing_failed"
    PROFILING_FAILED = "profiling_failed"

    # Error after entering the running state
    ERRORED = "errored"


class ProjectAssetT(str, ExtendedEnum):
    DATA = "data"
    MODEL = "model"


class ProjectAssetSourceT(str, ExtendedEnum):
    S3 = "s3"
    LOCAL = "local"
    LFS = "lfs"


class InstanceTypeT(str, ExtendedEnum):
    # Also change in taktile-services: services/deployment-api/app/deployment_api/core/t/konduit.py
    GP_SMALL = "gp.small"
    GP_MEDIUM = "gp.medium"
    GP_LARGE = "gp.large"
    GP_XLARGE = "gp.xlarge"
    GP_2XLARGE = "gp.xxlarge"
    GPU_SMALL = "gpu.small"
    CPU_SMALL = "cpu.small"
    CPU_MEDIUM = "cpu.medium"
    CPU_LARGE = "cpu.large"
    CPU_XLARGE = "cpu.xlarge"
    CPU_2XLARGE = "cpu.xxlarge"
    CPU_3XLARGE = "cpu.xxxlarge"


class EndpointKinds(str, ExtendedEnum):
    CUSTOM = "custom"
    TABULAR = "tabular"
    BINARY = "binary"
    REGRESSION = "regression"
    MULTICLASS = "multiclass"


class RestSchemaTypes(ExtendedEnum):
    DICT = "Dict"
    SEQUENCE = "Sequence"
    ARRAY = "Array"
    FLAT_ARRAY = "FlatArray"
    SERIES = "Series"
    DATAFRAME = "DataFrame"
    SINGLE_VALUE = "SingleValue"
    CUSTOM_MODEL = "CustomModel"
    ANY = "Any"


class AccessKind(str, ExtendedEnum):
    # see also corresponding AccessKind on t-api
    OWNER = "owner"
    VIEWER = "viewer"



import json
import sys
from typing import List

import click
import yaml
from pydantic import UUID4, BaseModel

from tktl.cli.common import (
    LOG,
    ClickGroup,
    filter_deploy_shared_options,
    filter_repo_shared_options,
    get_cmd_shared_options,
    to_uuid,
)
from tktl.commands import deployments as deployments_commands
from tktl.core.config import settings
from tktl.core.schemas.repository import Endpoint
from tktl.core.schemas.utils import get_table_from_list_of_schemas
from tktl.core.t import EndpointKinds
from tktl.login import validate_decorator


@click.group("get", help="Get resources", cls=ClickGroup, **settings.HELP_COLORS_DICT)
def get():
    pass


@get.command("repositories", help="Get repository resources")
@get_cmd_shared_options
@filter_repo_shared_options
@validate_decorator
def get_repositories(
    repo: str, owner: str, all_resources: bool, output: str, full: bool,
):
    command = deployments_commands.GetRepositories()
    result = command.execute(
        repository_name=repo, repository_owner=owner, return_all=all_resources,
    )
    keys = None
    if not full:
        keys = ["ID", "FULL NAME", "DEPLOYMENTS"]
        result = list(set(result))
    return produce_output(resources=result, keys=keys, maxwidth=200, output_kind=output)


@get.command("deployments", help="Get deployment resources")
@click.argument("repository_id", required=False, callback=to_uuid)
@get_cmd_shared_options
@filter_deploy_shared_options
@filter_repo_shared_options
@validate_decorator
def get_deployment_by_repo_id(
    output: str,
    repository_id: UUID4 = None,
    repo: str = None,
    owner: str = None,
    commit: str = None,
    branch: str = None,
    status: str = None,
    all_resources: bool = False,
    full: bool = None,
):
    command = deployments_commands.GetDeployments()
    if output == "stdout" and not repository_id:
        LOG.warning(
            "Fetching multiple deployments, this can take a few seconds. "
            "Pass a repository_id or branch/commit if you only want the deployments for a specific repository"
        )

    result = command.execute(
        repository_id=repository_id,
        repository_name=repo,
        repository_owner=owner,
        git_hash=commit,
        branch_name=branch,
        status_name=status,
        return_all=all_resources,
    )
    keys = ["ID", "BRANCH @ COMMIT", "STATUS", "ENDPOINTS"] if not full else None
    return produce_output(resources=result, keys=keys, maxwidth=350, output_kind=output)


@get.command("endpoints", help="Get endpoint resources")
@click.argument("deployment_id", required=False, callback=to_uuid)
@click.option("-e", "--endpoint", help="Endpoint name")
@click.option(
    "-k",
    "--endpoint-kind",
    help="Endpoint kind",
    type=click.Choice(EndpointKinds.list()),
)
@get_cmd_shared_options
@filter_deploy_shared_options
@filter_repo_shared_options
@validate_decorator
def get_endpoint_by_deployment_id(
    deployment_id: UUID4,
    repo: str,
    owner: str,
    commit: str,
    branch: str,
    status: str,
    endpoint_kind,
    endpoint: str,
    all_resources: bool,
    output: str,
    full: bool = True,
):
    command = deployments_commands.GetEndpoints()
    if output == "stdout" and not deployment_id:
        LOG.warning(
            "Fetching multiple endpoints, this can take a few seconds. "
            "Pass a deployment_id if you only want the endpoints for a specific deployment"
        )

    result = command.execute(
        deployment_id=deployment_id,
        endpoint_name=endpoint,
        endpoint_kind=endpoint_kind,
        repository_name=repo,
        repository_owner=owner,
        git_hash=commit,
        branch_name=branch,
        status_name=status,
        return_all=all_resources,
    )
    return produce_output(resources=result, output_kind=output, keys=None, maxwidth=130)


class GenericResourceList(BaseModel):
    __root__: List


def produce_output(
    resources: List[Endpoint],
    output_kind: str,
    keys: List[str] = None,
    maxwidth: int = None,
):
    if not resources:
        LOG.error("No resources found")
        sys.exit(1)
    if output_kind == "stdout":
        table = get_table_from_list_of_schemas(resources, maxwidth=maxwidth, keys=keys)
        LOG.log(table)
    elif output_kind == "json":
        LOG.log(GenericResourceList.parse_obj(resources).json())
    elif output_kind == "yaml":
        LOG.log(yaml.dump_all([json.loads(r.json()) for r in resources]))



import functools
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import pandas  # type: ignore
from cached_property import cached_property  # type: ignore

from tktl.core.exceptions import exceptions
from tktl.core.t import EndpointKinds
from tktl.core.utils import (
    DelayedLoaderX,
    DelayedLoaderXProfile,
    DelayedLoaderY,
    check_and_get_value,
)
from tktl.registration.schema import EndpointInputSchema, EndpointOutputSchema
from tktl.registration.validation import (
    input_pandas_representation,
    output_pandas_representation,
    validate_binary,
    validate_outputs,
    validate_shapes,
)


class Endpoint(ABC):
    KIND: EndpointKinds
    input_schema: EndpointInputSchema
    output_schema: EndpointOutputSchema
    explain_input_schema: Optional[EndpointInputSchema]
    delayed_loading: bool
    profile_columns: Optional[List[str]]

    def __init__(
        self,
        func: Callable,
        profile_columns: Optional[List] = None,
        disable_rest_validation: bool = True,
    ):
        self._func = func
        self.delayed_loading = False
        self.profile_columns = (
            [str(c) for c in profile_columns] if profile_columns is not None else None
        )
        self.disable_rest_validation = disable_rest_validation

    @cached_property
    def has_arrow_sample_data(self):
        return (
            self.input_schema.supports_arrow_sample_data
            and self.output_schema.supports_arrow_sample_data
        )

    @cached_property
    def has_rest_sample_data(self):
        if self.disable_rest_validation:
            return False
        else:
            return (
                self.input_schema.supports_rest_sample_data
                and self.output_schema.supports_rest_sample_data
            )

    @property
    def func(self):
        return self._func

    @property
    def series_func(self) -> Union[Any, pandas.Series]:
        name = self._func.__name__

        def to_series(x):
            return pandas.Series(self.func(x))

        to_series.__name__ = name
        return to_series

    @cached_property
    @abstractmethod
    def profiling_supported(self):
        return True

    @cached_property
    def input_names(self):
        return []

    @cached_property
    def output_names(self):
        return []


class TabularEndpoint(Endpoint):
    explain_input_schema: Optional[EndpointInputSchema]
    KIND: EndpointKinds = EndpointKinds.TABULAR

    def __init__(self, func, X, y, profile_columns, disable_rest_validation):
        super().__init__(
            func, profile_columns, disable_rest_validation=disable_rest_validation
        )

        if isinstance(X, str) and isinstance(y, str):
            self.delayed_loading = True

            x_value = DelayedLoaderX(path=X, label=y)
            y_value = DelayedLoaderY(path=X, label=y)

        else:
            x_value = X
            y_value = y

        self.input_schema = EndpointInputSchema(
            value=x_value, endpoint_kind=self.KIND, endpoint_name=self.func.__name__
        )
        self.output_schema = EndpointOutputSchema(
            value=y_value, endpoint_kind=self.KIND, endpoint_name=self.func.__name__
        )

        if self.profiling_supported:
            if not profile_columns:
                self.profile_columns = self.input_schema.names

            if self.delayed_loading:
                explain_input_value = DelayedLoaderXProfile(
                    path=X, label=y, columns=self.profile_columns
                )
            else:
                explain_input_value = self.input_schema.pandas_representation[
                    self.profile_columns
                ]

            self.explain_input_schema = EndpointInputSchema(
                value=explain_input_value,
                endpoint_kind=self.KIND,
                endpoint_name=f"explain__{self.func.__name__}",
            )
        else:
            self.explain_input_schema = None

    def _profiling_supported(self) -> Union[bool, pandas.Series]:
        if (
            not self.input_schema.pandas_convertible
            or not self.output_schema.pandas_convertible
        ):

            return False
        try:
            value = check_and_get_value(self.input_schema.value, first_batch=True)
            predictions = self.func(value)
        except Exception as e:
            raise exceptions.ValidationException(
                f"Function provided is unable to produce predictions from given sample values: {e}"
            )
        return predictions

    @cached_property
    def profiling_supported(self):
        predictions = self._profiling_supported()
        if predictions is False:
            return False
        else:
            try:
                x_frame = input_pandas_representation(value=self.input_schema.value)
                y_series = output_pandas_representation(value=self.output_schema.value)
            except ValueError:
                return False
            return validate_outputs(predictions) and validate_shapes(
                x_frame, y_series, predictions
            )

    @cached_property
    def input_names(self):
        return self.input_schema.names

    @cached_property
    def explain_input_names(self):
        return self.explain_input_schema.names if self.explain_input_schema else []

    @cached_property
    def output_names(self):
        return self.output_schema.names if self.explain_input_schema else []


class BinaryEndpoint(TabularEndpoint):
    KIND = EndpointKinds.BINARY

    def __init__(self, func, X, y, profile_columns, disable_rest_validation=False):
        super().__init__(
            func, X, y, profile_columns, disable_rest_validation=disable_rest_validation
        )

    @cached_property
    def profiling_supported(self):
        predictions = self._profiling_supported()
        if predictions is False:
            return False
        return super().profiling_supported and validate_binary(predictions)


class RegressionEndpoint(TabularEndpoint):
    KIND = EndpointKinds.REGRESSION

    def __init__(self, func, X, y, profile_columns, disable_rest_validation=False):
        super().__init__(
            func, X, y, profile_columns, disable_rest_validation=disable_rest_validation
        )

    @cached_property
    def profiling_supported(self):
        predictions = self._profiling_supported()
        if predictions is False:
            return False
        return super().profiling_supported and validate_outputs(predictions)


class CustomEndpoint(Endpoint):
    def get_input_and_output_for_profiling(
        self, **kwargs
    ) -> Tuple[pandas.DataFrame, pandas.Series]:
        raise ValueError("Custom endpoint does not have profiling enabled")

    KIND = EndpointKinds.CUSTOM

    def __init__(
        self,
        func,
        payload_model=None,
        response_model=None,
        X=None,
        y=None,
        disable_rest_validation=False,
    ):
        try:
            assert isinstance(func, Callable)
        except AssertionError:
            raise exceptions.ValidationException("Endpoint function is not a callable")

        models_defined = payload_model and response_model
        x_and_y_defined = (X is not None) and (y is not None)
        if not models_defined and not x_and_y_defined:
            disable_rest_validation = True
            payload_model = response_model = Union[Dict, List]

        if models_defined and x_and_y_defined:
            raise ValueError(
                "For custom endpoints, either define sample data or payload models"
            )

        super().__init__(func, disable_rest_validation=disable_rest_validation)

        self.input_schema = EndpointInputSchema(
            value=X,
            endpoint_kind=self.KIND,
            endpoint_name=self.func.__name__,
            user_defined_model=payload_model,
        )
        self.output_schema = EndpointOutputSchema(
            value=y,
            endpoint_kind=self.KIND,
            endpoint_name=self.func.__name__,
            user_defined_model=response_model,
        )

    @cached_property
    def profiling_supported(self):
        return False


class Tktl:
    def __init__(self):
        self.endpoints = []

    # This is the user-facing decorator for function registration
    def endpoint(
        self,
        func: Callable = None,
        kind: str = EndpointKinds.REGRESSION,
        X: Any = None,
        y: Any = None,
        payload_model=None,
        response_model=None,
        disable_rest_validation: bool = False,
        profile_columns: Optional[List] = None,
    ):
        """Register function as a Taktile endpoint

        Parameters
        ----------
        func : Callable, optional
            Function that describes the desired operation, by default None
        kind : str, optional
            Specification of endpoint type ("regression", "binary", "custom"),
            by default "regression"
        X : pd.DataFrame, optional
            Reference input dataset for testing func. Used when argument "kind"
            is set to "regression" or "binary", by default None.
        y : pd.Series, optional
            Reference output for evaluating func. Used when argument "kind"
            is set to "regression" or "binary", by default None.
        payload_model:
            Type hint used for documenting and validating payload. Used in
            custom endpoints only.
        response_model:
            Type hint used for documenting and validating response. Used in
            custom endpoints only.
        disable_rest_validation: bool
            If true, skip Rest schema validation
        profile_columns: List, optional
            List of column names for which to run profiling. Used in binary
            and regression endpoints only.

        Returns
        -------
        Callable
            Wrapped function
        """
        endpoint: Union[
            TabularEndpoint, RegressionEndpoint, BinaryEndpoint, CustomEndpoint
        ]
        if func is None:
            return functools.partial(
                self.endpoint,
                kind=kind,
                X=X,
                y=y,
                payload_model=payload_model,
                response_model=response_model,
                disable_rest_validation=disable_rest_validation,
                profile_columns=profile_columns,
            )
        if kind == "tabular":
            endpoint = TabularEndpoint(
                func=func,
                X=X,
                y=y,
                profile_columns=profile_columns,
                disable_rest_validation=disable_rest_validation,
            )
        elif kind == "regression":
            endpoint = RegressionEndpoint(
                func=func,
                X=X,
                y=y,
                profile_columns=profile_columns,
                disable_rest_validation=disable_rest_validation,
            )
        elif kind == "binary":
            endpoint = BinaryEndpoint(
                func=func,
                X=X,
                y=y,
                profile_columns=profile_columns,
                disable_rest_validation=disable_rest_validation,
            )
        elif kind == "custom":
            endpoint = CustomEndpoint(
                func=func,
                payload_model=payload_model,
                response_model=response_model,
                X=X,
                y=y,
                disable_rest_validation=disable_rest_validation,
            )
        else:
            raise exceptions.ValidationException(f"Unknown endpoint kind: '{kind}'")

        self.endpoints.append(endpoint)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            pred = func(*args, **kwargs)
            return pred

        return wrapper



import pandas as pd  # type: ignore

from tktl.future.utils import JSONStructure


def dataframe_deserialize(
    *, value: JSONStructure, sample: pd.DataFrame
) -> pd.DataFrame:
    return pd.DataFrame(value)


def dataframe_serialize(*, value: pd.DataFrame) -> str:
    return value.to_json(orient="records", date_format="iso")


def dataframe_to_example(*, value: pd.DataFrame) -> str:
    return value.iloc[[0]].to_json(orient="records", date_format="iso")



import pathlib
import typing as t

import pandas as pd  # type: ignore
import pyarrow.parquet as pq  # type: ignore

from tktl.core.loggers import LOG
from tktl.future.lazy_loading import lazify_parquet_metadata


def lazify_file(
    *,
    source_path: pathlib.Path,
    target_path: pathlib.Path,
    data: t.Optional[pd.DataFrame],
):
    LOG.trace(f"Reading file '{source_path}'...")
    table = pq.read_table(source_path)

    table = lazify_parquet_metadata(table, data)

    LOG.trace(f"Writing patched file to '{target_path}'...")
    pq.write_table(table, target_path)



from typing import Dict

from tktl.commands.validate import build_image
from tktl.core.exceptions.exceptions import MissingDocker
from tktl.core.loggers import LOG, stream_blocking_logs


def run_container(
    path: str,
    nocache: bool,
    background: bool,
    auth_enabled: bool,
    color_logs: bool,
    secrets: Dict = None,
):
    try:
        dm, image = build_image(path, nocache=nocache, secrets=secrets)
    except MissingDocker:
        LOG.log(
            "Couldn't locate docker, please make sure it is installed, or use the DOCKER_HOST environment variable",
            color="red",
        )
        return
    LOG.log("Waiting for service to start...")
    arrow_container, rest_container = dm.run_containers(
        image, detach=True, auth_enabled=auth_enabled
    )

    if background:
        return arrow_container, rest_container
    stream_blocking_logs(arrow_container, rest_container, color_logs=color_logs)
